unit tsxSceneManager;
////////////////////////////////////////////////////////
// tSX Builder Sample Component                       //
// TtsxSceneManager - trueSpace Scene Graph traversal //
//          control. Represents the entire scene as a //
//          hierarchical treeview.                    //
////////////////////////////////////////////////////////

interface

uses classes, windows, messages, tsxTypes, Controls, StdCtrls, ComCtrls, dialogs, SysUtils,
  tSXPluginHelper;

Type
  //Simple tsxapi object wrapper for display in scene graph
  PTtsxObjectNode = ^TtsxObjectNode;
  TtsxObjectNode = class(TObject)
  private
    FType : tsxSOBJTYPE;
    FSObj : TSXSOBJ;
  public
    property ObjectType : tsxSOBJTYPE read FType write FType;
    constructor Create(AType : tsxSOBJTYPE; AObj : tsxSObj);
    property SObj: tsxSObj read FSobj write FSobj;
    function ObjectCaption: string;
  end;

  //The Scene Graph component
  TtsxSceneManager = class(TTreeView)
  private
    FClickSelect: Boolean;
    procedure ClearAll;
    procedure ExpandTree(APossObj : TsxSObj; AParent: TTreeNode; AChild: Boolean);
    procedure SetClickSelect(Value: Boolean);
  protected
    procedure Deleting(Sender: TObject; Node: TTreeNode);
    procedure Loaded; override;
    procedure Click; override;
    constructor Create(AOwner: TComponent); override;
    function FindObject(AObject: TsxSObj; var ALocation: TTreeNode): Boolean;
    function AddChildNode(NewNode :TtsxObjectNode; Position : TTreeNode): TTreeNode;
    function AddSiblingNode(NewNode :TtsxObjectNode; Position : TTreeNode): TTreeNode;
  public
    procedure Select;
    procedure RefreshGraph;
  published
    property ClickSelect: Boolean read FClickSelect write SetClickSelect;
  end;

implementation


//********* TtsxObjectNode **********//
constructor TtsxObjectNode.Create(AType : tsxSOBJTYPE; AObj : tsxSObj);
//initialize the object info
begin
  FType := AType;
  FSObj := AObj;
end;

function TtsxObjectNode.ObjectCaption: string;
//Create a caption for display in the treeview outline for the object
var
  ObjText : string;
  pName : PChar;
  iNameRetSz : integer;
begin
  case ObjectType of
    e_tsxUNDEFINED : ObjText := 'Undefined - ';
    e_tsxMATERIAL  : ObjText := 'Material - ';
    e_tsxSCENE     : ObjText := 'Scene';
    e_tsxSELECTABLE: ObjText := 'Selectable - ';
    e_tsxAXES      : ObjText := 'Axes - ';
    e_tsxGNODETYPES: ObjText := 'Generic GNode - ';
    e_tsxCAMERA    : ObjText := 'Camera - ';
    e_tsxLIGHT     : ObjText := 'Light - ';
    e_tsxMNODETYPES: ObjText := 'Generic MNode - ';
    e_tsxPOLYHEDRON: ObjText := 'Polyhedron - ';
    e_tsxGROUP     : ObjText := 'Group - ';
    e_tsxLODGROUP  : ObjText := 'LOD Group - ';
    e_tsxMBALLOBJ  : ObjText := 'MetaBall Object - ';
    e_tsxSOBJEND   : ObjText := 'Invalid Object - ';
    end;
  if (tsxIsGNode(Sobj) = e_tsxTRUE) then
    begin
    //Get the objects name from trueSpace
    pName := AllocMem(SizeOf(char)*255);
    iNameRetSz := tsxGNodeGetName(Sobj, pName, 255);
    pName[iNameRetSz] := #0;
    ObjText := ObjText + StrPas(pName);
    FreeMem(pName, SizeOf(char)*255);
    end;
  Result := ObjText;
end;

//********* TtsxSceneManager ********//
procedure TtsxSceneManager.Deleting(Sender: TObject; Node: TTreeNode);
//Used for the OnDeletion event of the TTreeView, to deallocate the associated
//TtsxObjectNode object for this TreeView node
begin
  TtsxObjectNode(Node.Data).Free;
end;

constructor TtsxSceneManager.Create(AOwner: TComponent);
//initialization of the scene graph component
begin
  inherited Create(AOwner);
  FClickSelect := False;
  OnDeletion := Deleting;
end;

function TtsxSceneManager.FindObject(AObject: TsxSObj; var ALocation: TTreeNode): Boolean;
//Seeks for a particular TsxSObj in the treeview control
//Returns the node if the object is found, returns false if
//the object is not contained in the treeview nodes
var
  Index : integer;
begin
  result := False;
  for Index := 0 to (Items.Count - 1) do
    if (TtsxObjectNode(Items.Item[Index].Data).Sobj = AObject) then
      begin
      ALocation := Items.Item[Index];
      result := True;
      Exit;
      end;
end;

procedure TtsxSceneManager.Select;
//Sets the treeview selected node to the current scene's selected object
//if the object is contained in the treeview
var
  ALocation : TTreeNode;
begin
  RefreshGraph;
  if FindObject(tsxGetCurrentSelection, ALocation) then
    Selected := ALocation;
end;

procedure TtsxSceneManager.SetClickSelect(Value: Boolean);
//Private Accessor method for the ClickSelect property
begin
  if FClickSelect <> Value then
    FClickSelect := Value;
end;

procedure TtsxSceneManager.Click;
//This is the inherited Click event of the TTreeView. We will execute this
//code after the inherited code is executed (after the event handler fires)
var
  TheCurrentNode : Pointer;
begin
  inherited Click;
  if not (csDesigning in ComponentState) then //Make sure we are not in the IDE
    if (FClickSelect //Stop if ClickSelect property is false (we dont change
                     //the selection when a new item is clicked
      and (Selected <> nil) //Stop if there is nothing in the treeview to select
      and (Selected.Data <> nil) //Stop if the selected item in the treeview does not have an associated object
      and (TtsxObjectNode(Selected.Data).SObj <> nil)) then //Stop if the selected object does not have a pointer to a trueSpace object
        begin
        //If the selected item in the treeview is selectable by trueSpace,
        //Then make it the current selection
        TheCurrentNode := TtsxObjectNode(Selected.Data).SObj;
        if (tsxIsSelectable(TheCurrentNode) = e_tsxTRUE) then
          tsxSelectSobj(TheCurrentNode, e_tsxSELECT, e_tsxFALSE);
        end;
end;

procedure TtsxSceneManager.ClearAll;
//Clear all Node data (TtsxObjectNodes), and then clear all of the actual treeview nodes
var
  Index : integer;
begin
  for Index := 0 to (Items.Count - 1) do
    if Assigned(Items.Item[Index].Data) then
      TtsxObjectNode(Items.Item[Index].Data).Free;
  Self.Items.Clear;
end;

procedure TtsxSceneManager.RefreshGraph;
//Refresh the control by querying the tSX API
var
  tmpObject : TtsxObjectNode;
  tmpTsxObj : Pointer;
  LastNode : TTreeNode;
begin
  //Clear anything that is already displaying
  ClearAll;
  //Grab the scene node
  tmpTsxObj := tsxGetScene;
  //Make a graph node to represent the scene node, and add it as the root of the treeview
  tmpObject := TtsxObjectNode.Create(tsxSobjGetType(tmpTsxObj), tmpTsxObj);
  LastNode := AddChildNode(tmpObject, Items.GetFirstNode);
  //Retrieve the first object node, and use it as a starting point to recursively
  //load the entire scene hierarchy
  ExpandTree(tsxSceneGetFirstNode, LastNode, True);
  //Call the Expand method of TTreeView to expand all levels of the hierarchy
  FullExpand;
end;

procedure TtsxSceneManager.ExpandTree(APossObj : TsxSObj; AParent: TTreeNode; AChild: Boolean);
//Recursive method which builds the scene graph
//I don't feel like explaining recursion,... so try to figure it out :)
var
  ThisNode     : TtsxObjectNode;
  ThisPosition : TTreeNode;
  FirstChild   : TsxSobj;
  FirstSibling : TsxSobj;
begin
  if (tsxIsGNode(APossObj) = e_tsxTRUE) then
    begin
    ThisNode := TtsxObjectNode.Create(tsxSobjGetType(APossObj), APossObj);
    if AChild then
      ThisPosition := AddChildNode(ThisNode, AParent)
      else
      ThisPosition := AddSiblingNode(ThisNode, AParent);
    FirstChild := tsxGNodeGetFirstChild(ThisNode.SObj);
    ExpandTree(FirstChild, ThisPosition, True);
    FirstSibling := tsxGNodeGetNext(ThisNode.SObj);
    ExpandTree(FirstSibling, ThisPosition, False);
    end;
end;

function TtsxSceneManager.AddSiblingNode(NewNode : TtsxObjectNode; Position : TTreeNode): TTreeNode;
//Add a treenode at the same level as the node specified by Position
begin
  result := Items.AddObject(Position, NewNode.ObjectCaption, NewNode)
end;

function TtsxSceneManager.AddChildNode(NewNode : TtsxObjectNode; Position : TTreeNode): TTreeNode;
//Add a treenode as a child of the node specified by Position
begin
  result := Items.AddChildObject(Position, NewNode.ObjectCaption, NewNode)
end;

procedure TtsxSceneManager.Loaded;
//This is the inherited Loaded method of TTreeView. If the control is running
//in the Delphi IDE, we just add a simple node, otherwise we try to set the
//component's current selection to reflect the state of the trueSpace selection
begin
  inherited Loaded;
  if not (csDesigning in ComponentState) then {RefreshGraph} Select
    else Items.AddChild(Items.GetFirstNode, 'Scene Manager (design mode)');
end;

end.
